#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os


def add_metadata(t_file):
    """
    Adds useful metadata to the file.

    Added metadata:
    * line number
    """
    tmp_dir = os.path.dirname(t_file)
    filename = os.path.basename(t_file)
    tmp_out = os.path.join(tmp_dir, filename + ".metadata.tmp")

    with open(tmp_out, "w") as fout:
        with open(t_file, "r") as fin:
            line = fin.readline()
            i = 1
            while line != "":
                metaline = "@!meta:{}!@".format(i) + line
                fout.write(metaline)
                i += 1
                line = fin.readline()

    return tmp_out


def remove_comments(t_file):
    """
    Removes comments from the file.
    """
    tmp_dir = os.path.dirname(t_file)
    filename = os.path.basename(t_file)
    tmp_out = os.path.join(tmp_dir, filename + ".nocomments.tmp")

    with open(tmp_out, "w") as fout:
        with open(t_file, "r") as fin:
            line = fin.readline()
            while line != "":
                not_commented = line.split("#")[0]
                fout.write(not_commented.strip())

                if not_commented.strip() != "":
                    fout.write("\n")

                line = fin.readline()

    return tmp_out


def remove_params_newline(t_file):
    """
    Removes newlines on parameters.
    """

    tmp_dir = os.path.dirname(t_file)
    filename = os.path.basename(t_file)
    tmp_out = os.path.join(tmp_dir, filename + ".nonewlines.tmp")

    with open(tmp_out, "w") as fout:
        with open(t_file, "r") as fin:
            line = fin.readline()
            while line != "":
                if line.strip().endswith("\\"):
                    nonewline = line.strip().replace("\\", " ")
                    fout.write(nonewline)
                else:
                    fout.write(line.strip() + "\n")

                line = fin.readline()

    return tmp_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
BLIFPARSER: A simple parser for BLIF files
"""

__author__ = "Zenaro Stefano"

import os
import shutil
import re
import tempfile

try:
    from ._version import __version__  # noqa: F401

    from . import keywords
    from . import utils
    from . import graph

except (ImportError, ModuleNotFoundError):
    from _version import __version__  # noqa: F401

    import utils
    import keywords
    import graph


class BlifParser:

    def __init__(self, t_file):  # noqa: C901
        """
        Parses the <t_file> BLIF file.

        :param str t_file: input BLIF file
        """
        # prepare the input file
        prepared_file = self.prepare_file(t_file)

        self.blif = keywords.generic.Blif()

        is_boolfunc = False
        boolfunc_dontcare = False
        is_fsm = False
        is_model = False

        with open(prepared_file) as fin:
            line = fin.readline()
            i = 1

            while line != "":
                linestrip = line.strip()

                # get and remove the metadata from the current line
                metadata = re.match("@!(.*?)!@", linestrip)
                if metadata:
                    i = metadata.groups()[0].lstrip("meta:")

                linestrip = re.sub("@!(.*?)!@", "", linestrip)

                keyword = None
                params = None

                # get the keyword in the current line
                if linestrip.startswith("."):
                    is_boolfunc = False
                    keyword = linestrip.split(" ")[0].strip()

                    try:
                        params = " ".join(linestrip.split(" ")[1:]).strip()
                    except IndexError:
                        pass

                try:
                    if linestrip == "":
                        # skip empty lines
                        pass
                    elif keyword:
                        # found a keyword
                        if keyword == ".model":
                            is_model = True
                            self.blif.model = keywords.generic.Model(params)
                        elif keyword == ".inputs":
                            self.blif.inputs = keywords.generic.Inputs(params)
                        elif keyword == ".outputs":
                            self.blif.outputs = keywords.generic.Outputs(params)
                        elif keyword == ".search":
                            self.blif.imports.append(keywords.subfiles.Search(params))
                        elif keyword == ".names":
                            is_boolfunc = True
                            self.blif.booleanfunctions.append(keywords.generic.Names(params, boolfunc_dontcare))
                        elif keyword == ".latch":
                            self.blif.latches.append(keywords.generic.Latch(params))
                        elif keyword == ".subckt":
                            self.blif.subcircuits.append(keywords.subfiles.Subckt(params))
                        elif keyword == ".start_kiss":
                            self.blif.fsm.ispresent = True
                            is_fsm = True
                        elif keyword == ".i":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected .i keyword: "
                                                          "needs to be between .start_kiss and .end_kiss keywords".format(i))
                            self.blif.fsm.i = keywords.fsm.I(params)
                        elif keyword == ".o":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected .o keyword: "
                                                          "needs to be between .start_kiss and .end_kiss keywords".format(i))
                            self.blif.fsm.o = keywords.fsm.O(params)
                        elif keyword == ".s":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected .s keyword: "
                                                          "needs to be between .start_kiss and .end_kiss keywords".format(i))
                            self.blif.fsm.s = keywords.fsm.S(params)
                        elif keyword == ".p":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected .p keyword: "
                                                          "needs to be between .start_kiss and .end_kiss keywords".format(i))
                            self.blif.fsm.p = keywords.fsm.P(params)
                        elif keyword == ".r":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected .r keyword: "
                                                          "needs to be between .start_kiss and .end_kiss keywords".format(i))
                            self.blif.fsm.r = keywords.fsm.R(params)
                        elif keyword == ".end_kiss":
                            if not is_fsm:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected end of fsm: "
                                                          "there needs to be a .start_kiss keyword "
                                                          "BEFORE the .end_kiss keyword".format(i))
                            is_fsm = False
                        elif keyword == ".exdc":
                            boolfunc_dontcare = True
                        elif keyword == ".code":
                            self.blif.fsm.statecodes.append(keywords.fsm.Code(params))
                        elif keyword == ".end":
                            if not is_model:
                                self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected end of model on line {}".format(i))
                            is_model = False

                        try:
                            self.blif.nkeywords[keyword] += 1
                        except KeyError:
                            self.blif.problems.append("[ERROR][LINE ~ {}] Invalid keyword: '{}'".format(i, keyword))

                    else:
                        # found transition/truth table
                        if is_boolfunc:
                            curr_boolfunc = self.blif.booleanfunctions[-1]
                            curr_boolfunc.truthtable.append([char for char in linestrip if char != " "])

                        elif is_fsm:
                            fsm = self.blif.fsm
                            fsm.transtable.append([el for el in linestrip.split(" ") if el != " "])

                        elif linestrip != "":
                            self.blif.problems.append("[ERROR][LINE ~ {}] Unexpected text: '{}'".format(i, linestrip))

                except Exception as e:
                    self.blif.problems.append("[PARSING ERROR][LINE ~ {}] ".format(i) + str(e))

                line = fin.readline()

        # if an FSM is present in the file, check if it is valid
        if self.blif.fsm.ispresent:
            try:
                self.blif.fsm.is_valid()
            except Exception as e:
                self.blif.problems.append("[FSM PROBLEM] " + str(e))

        # check if each boolean function is valid
        for boolfunc in self.blif.booleanfunctions:
            try:
                boolfunc.is_valid()
            except Exception as e:
                self.blif.problems.append("[BOOLEAN FUNCTION PROBLEM] " + str(e))

    def prepare_file(self, t_file):
        """
        Prepares the <t_file> file for parsing.

        First it makes a copy of the file in a temporary folder,
        then adds metadata in the file (line number),
        removes the comments and the newlines created with "\".

        :param str t_file: input file path
        :return str nonewlines: "ready to be parsed" file path
        """
        filepath = os.path.abspath(t_file)

        # create a copy of the file inside a temporary folder
        tmp_dir = tempfile.mkdtemp()
        tmp_filepath = os.path.join(tmp_dir, "example.blif")
        shutil.copyfile(filepath, tmp_filepath)

        # adds metadata, removes comments and newlines on keywords' params
        metadata = utils.add_metadata(tmp_filepath)
        nocomments = utils.remove_comments(metadata)
        nonewlines = utils.remove_params_newline(nocomments)

        return nonewlines

    def get_graph(self):
        return graph.parse_blif(self.blif)


def main():
    import sys

    print("")

    if len(sys.argv) == 2:
        filepath = os.path.abspath(sys.argv[1])
        blif = BlifParser(filepath).blif

        print("ISSUES LIST:\n")

        for problem in blif.problems:
            print(problem)

        print("")
        print("=" * 50)
        print("\nREPORT:\n")
        print("* {} issues found".format(len(blif.problems)))
    else:
        print("blifparser expects only one parameter: the input BLIF file path")


if __name__ == "__main__":
    main()





class Fsm:
    def __init__(self):
        """
        Defines an Fsm() object.

        Attributes:
        * self.i: represents the .i keyword (number of inputs)
        * self.o: represents the .o keyword (number of outputs)
        * self.s: represents the .s keyword (number of unique states)
        * self.p: represents the .p keyword (number of state transitions)
        * self.r: represents the .r keyword (name of the reset state)
        * self.ispresent: True if an FSM is present in the blif file
        * self.transtable: list of lists (each row is a state transition)
            > Each row contains 4 elements: ["<inputs>", "<current_state>", "<next_state>", "<outputs>"]
        * self.statecodes: list of Code() objects (they represent .code keywords, state encodings)
        """
        self.i = None
        self.o = None
        self.s = None
        self.p = None
        self.r = None
        self.ispresent = False
        self.transtable = []
        self.statecodes = []

    def is_valid(self):
        """
        Validates all the elements present in an Fsm() object.
        """
        # Be sure that self.ispresent is kept boolean
        if not isinstance(self.ispresent, bool):
            raise TypeError("Something went wrong: self.ispresent should be boolean")

        # check if input and output are specified
        self.validate_i()
        self.validate_o()

        # Check transition table
        self.validate_transtable()

        # Check .p, .s, .r keywords
        self.validate_p()
        self.validate_s()
        self.validate_r()

        # Check .code keywords
        self.validate_codes()

        return True

    def validate_i(self):
        """
        Validates .i keyword: it needs to be set as an I() instance.
        """
        if self.i:
            if not isinstance(self.i, I):
                raise TypeError(".i must be I() instance")
        else:
            raise ValueError(".i must be set")

    def validate_o(self):
        """
        Validates .o keyword: it needs to be set as an O() instance.
        """
        if self.o:
            if not isinstance(self.o, O):
                raise TypeError(".o must be O() instance")
        else:
            raise ValueError(".o must be set")

    def validate_p(self):
        """
        Validates .p keyword: it is valid if it is a P() instance or if it is None.

        If it is set its parameter needs to be equal to the number of rows
        present in the transition table.
        """
        # if .p is not None check its instance
        if self.p:
            if not isinstance(self.p, P):
                raise TypeError(".p must be P() instance")

            # be sure that the num_terms are equal to the
            # number of rows of the transition table
            num_terms = int(self.p.num)
            if len(self.transtable) != num_terms:
                raise ValueError(".p <num-terms> has incorrect <num-terms> value "
                                 "(it should be '{}' instead of '{}' because the "
                                 "transition table has that amount of rows)".format(len(self.transtable), num_terms))

    def validate_s(self):
        """
        Validates .s keyword: it is valid if it is a S() instance or if it is None.

        If it is set its parameter needs to be equal to the number of unique states
        present in the transition table.
        """
        state_names = self.get_state_names()

        if self.s:
            if not isinstance(self.s, S):
                raise TypeError(".s must be S() instance")

            num_states = int(self.s.num)
            expected_num = len(state_names)
            if expected_num != num_states:
                raise ValueError(".s <num-states> has incorrect <num-states> value "
                                 "(it should be '{}' instead of '{}' because the transition "
                                 "table contains that amount of unique states)".format(expected_num, num_states))

    def validate_r(self):
        """
        Validates .r keyword: it is valid if it is a R() instance or if it is None.

        If it is set its parameter needs to be a state that is
        present in the transition table.
        """
        state_names = self.get_state_names()

        if self.r:
            if not isinstance(self.r, R):
                raise TypeError(".r must be R() instance")

            reset_state = self.r.name
            if reset_state not in state_names:
                raise ValueError(".r <reset-state> has incorrect <reset-state> value "
                                 "(can't find '{}' in the transition table)".format(reset_state))

    def validate_codes(self):
        """
        Validates .code keywords: it is valid if self.statecodes contains Code() instances.
        """
        if not isinstance(self.statecodes, list):
            raise TypeError("Something went wrong: statecodes is not a list")

        # be sure that self.statecodes only contain states
        for code in self.statecodes:
            if not isinstance(code, Code):
                raise TypeError(".code in statecodes must be Code() instance")

    def get_state_names(self):
        """
        Get unique state names from the transition table.
        """
        states = set()
        for row in self.transtable:
            if len(row) == 4:
                states.add(row[1])
                states.add(row[2])

        return states

    def validate_transtable(self):  # noqa: C901
        """
        Validates the transition table.

        The transition table:
        * needs to contain lists of strings (4 for each list)
        * be sure that the first element (inputs) is make of 0 1 -
        * be sure that the inputs are the amount specified by the .i keyword
        * be sure that the last element (outputs) is make of 0 1 -
        * be sure that the outputs are the amount specified by the .o keyword
        """

        if not isinstance(self.transtable, list):
            raise TypeError("Something went wrong: transition table is not a list")

        for row in self.transtable:
            # be sure that it contains lists of strings
            if not isinstance(row, list):
                raise TypeError("'{}' is not a list".format(row))

            for el in row:
                if not isinstance(el, str):
                    raise TypeError("'{}' in '{}' is not a string".format(el, row))

            # be sure that there are four strings inside each list
            if len(row) != 4:
                raise ValueError("'{}' row of the transition table doesn't have all the "
                                 "expected data (<input> <current-state> <next-state> <output>)".format(" ".join(row)))

            # be sure that the first element (inputs) is make of 0 1 -
            for el in row[0]:
                if el not in ["0", "1", "-"]:
                    raise ValueError("'{}' input of the '{}' row (transition table) "
                                     "has unexpected characters (accepted chars are "
                                     "['0', '1', '-'])".format(row[0], " ".join(row)))

            # be sure that the inputs are the amount specified by the .i keyword
            if len(row[0]) != int(self.i.num):
                raise ValueError("'{}' row (transition table) has an unexpected number "
                                 "of inputs (found {} elements in '{}' but expected {} "
                                 "based on the .i parameter)".format(" ".join(row), len(row[0]), row[0], self.i.num))

            # be sure that the last element (outputs) is make of 0 1 -
            for el in row[3]:
                if el not in ["0", "1", "-"]:
                    raise ValueError("'{}' output of the '{}' row (transition table) "
                                     "has unexpected characters (accepted chars are "
                                     "['0', '1', '-'])".format(row[3], " ".join(row)))

            # be sure that the outputs are the amount specified by the .o keyword
            if len(row[3]) != int(self.o.num):
                raise ValueError("'{}' row (transition table) has an unexpected number "
                                 "of outputs (found {} elements in '{}' but expected {} "
                                 "based on the .o parameter)".format(" ".join(row), len(row[3]), row[3], self.o.num))

    def __str__(self):
        """Printed string."""
        # be sure that the object is validated
        self.is_valid()

        # adds .start_kiss, .i, .o keywords for printing
        fsm = ".start_kiss\n"
        fsm += self.i.__str__() + "\n"
        fsm += self.o.__str__() + "\n"

        # adds .p, .s, .r keywords (if not None) for printing
        if self.p:
            fsm += self.p.__str__() + "\n"

        if self.s:
            fsm += self.s.__str__() + "\n"

        if self.r:
            fsm += self.r.__str__() + "\n"

        fsm += "\n"

        # adds the transition table for printing
        for row in self.transtable:
            fsm += " ".join(row) + "\n"

        fsm += "\n"

        # adds .end_kiss keyword for printing
        fsm += ".end_kiss\n"

        # adds the states encodings for printing
        for state in self.statecodes:
            fsm += state.__str__() + "\n"

        return fsm


class I:
    def __init__(self, params):
        """
        Defines .i keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to be numeric (contain numbers)
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        try:
            int(params.strip())
        except ValueError:
            raise ValueError(".i parameter needs to be numeric (it must be made of digits 0-9)")

        self.num = params.strip()

    def __repr__(self):
        """Object representation."""
        return "I('" + self.num + "')"

    def __str__(self):
        """Printed string."""
        return ".i " + self.num


class O:
    def __init__(self, params):
        """
        Defines .o keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to be numeric (contain numbers)
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        try:
            int(params.strip())
        except ValueError:
            raise ValueError(".o parameter needs to be numeric (it must be made of digits 0-9)")

        self.num = params.strip()

    def __repr__(self):
        """Object representation."""
        return "O('" + self.num + "')"

    def __str__(self):
        """Printed string."""
        return ".o " + self.num


class S:
    def __init__(self, params):
        """
        Defines .s keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to be numeric (contain numbers)
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        try:
            int(params.strip())
        except ValueError:
            raise ValueError(".s parameter needs to be numeric (it must be made of digits 0-9)")

        self.num = params.strip()

    def __repr__(self):
        """Object representation."""
        return "S('" + self.num + "')"

    def __str__(self):
        """Printed string."""
        return ".s " + self.num


class P:
    def __init__(self, params):
        """
        Defines .p keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to be numeric (contain numbers)
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        try:
            int(params.strip())
        except ValueError:
            raise ValueError(".p parameter needs to be numeric (it must be made of digits 0-9)")

        self.num = params.strip()

    def __repr__(self):
        """Object representation."""
        return "P('" + self.num + "')"

    def __str__(self):
        """Printed string."""
        return ".p " + self.num


class R:
    def __init__(self, params):
        """
        Defines .r keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to contain one parameter with no spaces
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        self.name = params.strip()

        if " " in self.name or self.name == "":
            raise ValueError(".r accepts (and needs) only one parameter (the parameter can't contain spaces)")

    def __repr__(self):
        """Object representation."""
        return "R('" + self.name + "')"

    def __str__(self):
        """Printed string."""
        return ".r " + self.name


class Code:
    def __init__(self, params):
        """
        Defines .code keyword object.

        Validation steps:
        * params needs to be a string
        * params needs to contain two parameters separated by space(s)
        * the second parameter needs to contain zeros and ones only
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        v_params = [param for param in params.split(" ") if param != ""]

        if len(v_params) != 2:
            raise ValueError(".code expects two parameters")

        self.state_name = v_params[0]
        self.state_encoding = v_params[1]

        for el in self.state_encoding:
            if el not in ["0", "1"]:
                raise ValueError("Unexpected char in the state encoding "
                                 "(only '0's and '1's are accepted): "
                                 ".code " + self.state_name + " " + self.state_encoding)

    def __repr__(self):
        """Object representation."""
        return "Code('" + self.state_name + " " + self.state_encoding + "')"

    def __str__(self):
        """Printed string."""
        return ".code " + self.state_name + " " + self.state_encoding



import networkx as nx

try:
    from . import blifparser
except (ImportError, ModuleNotFoundError):
    import blifparser


class Graph:
    def __init__(self, nx_graph, longest_label, max_inputs):
        """
        Memorizes the networkx graph, 
        the longest_label length, and the max_inputs number.
        """
        self.nx_graph = nx_graph
        self.longest_label = longest_label
        self.max_inputs = max_inputs


class Node:
    progressive_id = 1  # makes sure that each id is unique
    def __init__(self):
        """
        Defines a node.

        A node has:
        * a type: input, output, boolean_function, latch
        * an id: identifies the node in a graph
        > Note: the id is a progressive integer. 
        > The variable that controls that progressive integer is shared between Node instances.

        A node can have:
        * many inputs
        * many outputs

        """
        self.inputs = []
        self.outputs = []
        self.type = None
        self.id = Node.progressive_id
        self.node_color = "black"

        Node.progressive_id += 1  # makes sure that each id is unique
    
    def __str__(self) -> str:
        """
        Returns the Node's id as a string
        """
        return str(self.id)


def parse_blif(t_blif):
    """
    Parses a Blif obect to create a graph.
    """
    # prepare nodes objects
    nodes = make_nodes(t_blif)

    # add the nodes to a directed graph
    G = nx.DiGraph()
    G.add_nodes_from(nodes)

    # connect nodes together and keep some statistics
    longest_label = 1
    max_inputs = 1
    for n1 in nodes:
        for n2 in nodes:
            for n1_output in n1.outputs:
                for n2_input in n2.inputs:
                    if n1_output == n2_input:
                        # n1 --> n2
                        G.add_edge(n1, n2, label=n1_output)

                # save stats to scale the image
                if len(n1.inputs) > max_inputs:
                    max_inputs = len(n1.inputs)

                if len(n1_output) > longest_label:
                    longest_label = len(n1_output)

    return Graph(G, longest_label, max_inputs)


def make_nodes(t_blif):
    """
    Creates nodes that are not binded to each other
    but with the necessary information to bind them later.

    TODO: maybe divide this into 5 functions? (one for each type of node)
    """
    nodes = []

    # blif inputs (.inputs) are nodes that have one output: the input value
    for input in t_blif.inputs.inputs:
        n = Node()
        n.outputs.append(input)
        n.type = "input"
        n.parent = None
        n.name = str(n.id)
        n.node_color = "red"
        nodes.append(n)
    
    # blif outputs (.outputs) are nodes that have one input: the output value
    for output in t_blif.outputs.outputs:
        n = Node()
        n.inputs.append(output)
        n.type = "output"
        n.parent = None
        n.name = str(n.id)
        n.node_color = "blue"
        nodes.append(n)

    # blif boolean functions, defined using .names arg1 arg2 ...
    # have one output and might have multiple inputs
    for function in t_blif.booleanfunctions:
        n = Node()

        # collect the inputs
        for input in function.inputs:
            n.inputs.append(input)
        
        n.outputs.append(function.output)
        n.type = "boolean_function"
        n.parent = None
        n.name = str(n.id)
        nodes.append(n)
    
    # blif latches (.latch) have one input and one output
    for latch in t_blif.latches:
        n = Node()
        n.inputs.append(latch.input)
        n.outputs.append(latch.output)
        n.type = "latch"
        n.parent = None
        n.name = str(n.id)
        nodes.append(n)
    
    # blif sub-circuits (.subckt) might have multiple inputs and/or multiple outputs.
    # > to distinguish inputs from outputs it is necessary to parse imported (.search) modules
    for subckt in t_blif.subcircuits:
        n = Node()
        
        # loop for each imported file
        for imported_blif in t_blif.imports:
            filepath = imported_blif.filepath
            parser = blifparser.BlifParser(filepath)
            subckt_data = parser.blif

            # check if we have found the .model referenced by .subckt
            if subckt.modelname == subckt_data.model.name:
                # loop for each input (.inputs) of the .model inside the .search-ed file
                for model_input in subckt_data.inputs.inputs:
                    # loop for each parameter of .subckt
                    for subckt_param in subckt.params:
                        subckt_input = subckt_param.split("=")[0]
                        # if the .subckt parameter is also a .model input (of the imported file)
                        # that means that the parameter is an input of the sub-circuit
                        if model_input == subckt_input:
                            n.inputs.append(subckt_param.split("=")[1])
                
                # loop for each output (.outputs) of the .model inside the .search-ed file
                for model_output in subckt_data.outputs.outputs:
                    # loop for each parameter of .subckt
                    for subckt_param in subckt.params:
                        subckt_output = subckt_param.split("=")[0]
                        # if the .subckt parameter is also a .model output (of the imported file)
                        # that means that the parameter is an output of the sub-circuit
                        if model_output == subckt_output:
                            n.outputs.append(subckt_param.split("=")[1])
        
        n.type = "subckt"
        n.parent = None
        n.name = str(n.id)
        nodes.append(n)

    return nodes


if __name__ == "__main__":
    a = Node()
    print("A: ", a.__dict__)

    b = Node()
    print("B: ", b.__dict__)
    print("A: ", a.__dict__)

    c = Node()
    print("C: ", c.__dict__)
    print("B: ", b.__dict__)
    print("A: ", a.__dict__)





class Search:

    def __init__(self, param):
        if not isinstance(param, str):
            raise TypeError("'{}' is not a string".format(param))

        self.filepath = param.strip()

        if self.filepath == "":
            raise ValueError(".search needs one parameter")

    def __repr__(self):
        """Object representation."""
        return "Search('" + self.filepath + "')"

    def __str__(self):
        """Printed string."""
        return ".search " + self.filepath


class Subckt:
    def __init__(self, params):
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        v_params = [param for param in params.split(" ") if param != ""]

        if len(v_params) < 2:
            raise ValueError(".subckt expects at least two parameters")

        self.modelname = v_params[0]
        self.params = v_params[1:]

        for param in self.params:
            if "=" not in param:
                raise ValueError("'{}' parameter is incorrect (there needs to be an equal sign '=')".format(param))

    def __repr__(self):
        """Object representation."""
        return "Subckt('" + self.modelname + " " + " ".join(self.params) + "')"

    def __str__(self):
        """Printed string."""
        return ".subckt " + self.modelname + " " + " ".join(self.params)



try:
    from . import fsm
except (ModuleNotFoundError, ImportError):
    import fsm


class Model:
    def __init__(self, modelname):
        """
        Defines a .model keyword object.

        Validation checks:
        * modelname needs to be a string
        * modelname needs to contain something (spaces are not accepted)
        """
        if not isinstance(modelname, str):
            raise TypeError("'{}' is not a string".format(modelname))

        self.name = modelname.strip()

        if " " in self.name or self.name == "":
            raise ValueError(".model accepts (and needs) only one parameter (the parameter can't contain spaces)")

    def __repr__(self):
        """Object representation."""
        return "Model('" + self.name + "')"

    def __str__(self):
        """Printed string."""
        return ".model " + self.name


class Inputs:
    def __init__(self, inputstring):
        """
        Defines a .inputs keyword object.

        Validation checks:
        * inputstring needs to be a string
            > inputs are separated by spaces
        """
        if not isinstance(inputstring, str):
            raise TypeError("'{}' is not a string".format(inputstring))

        self.inputs = [i for i in inputstring.split(" ") if i != ""]

        if len(self.inputs) == 0:
            raise ValueError(".inputs keyword expects at least one parameter")

    def __repr__(self):
        """Object representation."""
        return "Inputs('" + " ".join(self.inputs) + "')"

    def __str__(self):
        """Printed string."""
        return ".inputs " + " ".join(self.inputs)


class Outputs:
    def __init__(self, outputstring):
        """
        Defines a .outputs keyword object.

        Validation checks:
        * outputstring needs to be a string
            > outputs are separated by spaces
        """
        if not isinstance(outputstring, str):
            raise TypeError("'{}' is not a string".format(outputstring))

        self.outputs = [i for i in outputstring.split(" ") if i != ""]

        if len(self.outputs) == 0:
            raise ValueError(".outputs keyword expects at least one parameter")

    def __repr__(self):
        """Object representation."""
        return "Outputs('" + " ".join(self.outputs) + "')"

    def __str__(self):
        """Printed string."""
        return ".outputs " + " ".join(self.outputs)


class Names:
    def __init__(self, params, dontcare):
        """
        Defines a .names keyword object.

        Validation checks:
        * params needs to be a string
            > the last parameter in the string is the output
        * dontcare needs to be a boolean

        If dontcare is true, the output
        represents a don't care.
        """
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        if not isinstance(dontcare, bool):
            raise TypeError("'{}' is not a boolean".format(dontcare))

        self.v_params = [param for param in params.split(" ") if param != ""]

        self.inputs = []
        self.truthtable = []
        self.output = None
        self.is_dontcare = dontcare

        if len(self.v_params) > 1:
            self.inputs = self.v_params[:-1]

        self.output = self.v_params[-1]

    def is_valid(self):  # noqa: C901
        """
        Validates data in the Names() object.

        Validation steps:

        """
        # be sure that self.inputs is a list of strings
        if not isinstance(self.inputs, list):
            raise TypeError("Something went wrong: self.inputs should be a list")

        for el in self.inputs:
            if not isinstance(el, str):
                raise TypeError("Something went wrong: '{}' self.inputs element should be a string".format(el))

        # be sure that self.output is a string
        if not isinstance(self.output, str):
            raise TypeError("Something went wrong: '{}' self.output should be a string".format(self.output))

        # be sure that self.is_dontcare stays a boolean
        if not isinstance(self.is_dontcare, bool):
            raise TypeError("Something went wrong: '{}' self.is_dontcare is not a boolean".format(self.is_dontcare))

        # validate the truth table
        if not isinstance(self.truthtable, list):
            raise TypeError("Something went wrong: self.truthtable should be a list")

        expected_el_num = len(self.inputs) + 1

        for row in self.truthtable:
            if not isinstance(row, list):
                raise TypeError("row '{}' is not a list (under '{}')".format(row, self.__str__()))

            for el in row:
                if not isinstance(el, str):
                    raise TypeError("'{}' element is not a string (in '{}' under '{}')".format(el, row, self.__str__()))

            formatted_row = "".join(row[:-1]) + " " + row[-1]
            if len(row) != expected_el_num:
                raise ValueError(
                    "'{}' row should have {} inputs + 1 output: found {} instead "
                    "(under '{}')".format(
                        formatted_row,
                        len(self.inputs),
                        len(row),
                        self.__str__()
                    )
                )

            for el in row[:-1]:
                if el not in ["0", "1", "-"]:
                    raise ValueError("Found unexpected char '{}' as input in row '{}' "
                                     "(under '{}'), only '1', '0' and '-' "
                                     "are accepted".format(el, formatted_row, self.__str__()))

            if row[-1] not in ["0", "1"]:
                raise ValueError("Found unexpected char '{}' as output in row '{}' "
                                 "(under '{}'), only '1' and '0' "
                                 "are accepted".format(el, formatted_row, self.__str__()))

        return True

    def __repr__(self):
        """Object representation."""
        return "Names('" + " ".join(self.inputs) + " " + self.output + "', " + str(self.is_dontcare) + ")"

    def __str__(self):
        """Printed string."""
        names = ""
        if self.is_dontcare:
            names = ".exdc\n"

        names += ".names " + " ".join(self.inputs) + " " + self.output

        if len(self.truthtable) > 0:
            names += "\n"
            for row in self.truthtable:
                formatted_row = "".join(row[:-1]) + " " + row[-1]
                names += formatted_row + "\n"

        return names


class Latch:
    def __init__(self, params):  # noqa: C901
        if not isinstance(params, str):
            raise TypeError("'{}' is not a string".format(params))

        self.v_params = [param for param in params.split(" ") if param != ""]

        self.problems = []
        self.input = None
        self.output = None
        self.type = None
        self.control = None
        self.initval = None

        # set the correct attributes based on the number of parameters
        if len(self.v_params) < 2:
            raise Exception("You need to specify at least an input and an output")

        elif len(self.v_params) == 2:
            self.problems.append(
                "WARNING: you should specify the initial value "
                "(otherwise you'll need to set it later using the set_state command)"
            )

        elif len(self.v_params) == 3:
            self.initval = self.v_params[2]

        elif len(self.v_params) == 4:
            self.type = self.v_params[2]
            self.control = self.v_params[3]

        elif len(self.v_params) == 5:
            self.type = self.v_params[2]
            self.control = self.v_params[3]
            self.initval = self.v_params[4]

        elif len(self.v_params) > 5:
            raise Exception("Too many parameters (correct usage is: .latch <input> <output> [<type> <control>] [<init-val>])")

        # set as values the parameters that must be specified
        self.input = self.v_params[0]
        self.output = self.v_params[1]

        # check if parameters have correct values

        if self.type:
            if self.type not in ["fe", "re", "ah", "al", "as"]:
                raise ValueError("<type> should be one of these values: ['fe', 're', 'ah', 'al', 'as']")

        if self.initval:
            if self.initval not in ["0", "1", "2", "3"]:
                raise ValueError("<init-val> should be one of these values: ['0', '1', '2', '3']")

    def __repr__(self):
        """Object representation."""
        latch = "Latch('"

        if len(self.v_params) == 2:
            latch += self.input + " " + self.output
        elif len(self.v_params) == 3:
            latch += self.input + " " + self.output + " " + self.initval
        elif len(self.v_params) == 4:
            latch += self.input + " " + self.output + " " + self.type + " " + self.control
        else:
            latch += self.input + " " + self.output + " " + self.type + " " + self.control + " " + self.initval

        latch += "')"
        return latch

    def __str__(self):
        """Printed string."""
        latch = ".latch "

        if len(self.v_params) == 2:
            latch += self.input + " " + self.output
        elif len(self.v_params) == 3:
            latch += self.input + " " + self.output + " " + self.initval
        elif len(self.v_params) == 4:
            latch += self.input + " " + self.output + " " + self.type + " " + self.control
        else:
            latch += self.input + " " + self.output + " " + self.type + " " + self.control + " " + self.initval

        return latch


class Blif:
    def __init__(self):
        self.model = None
        self.inputs = None
        self.outputs = None
        self.fsm = fsm.Fsm()
        self.imports = []
        self.subcircuits = []
        self.latches = []
        self.booleanfunctions = []
        self.problems = []

        self.nkeywords = {
            ".model": 0,
            ".inputs": 0,
            ".outputs": 0,
            ".search": 0,
            ".subckt": 0,
            ".latch": 0,
            ".names": 0,
            ".end": 0,
            ".start_kiss": 0,
            ".i": 0,
            ".o": 0,
            ".s": 0,
            ".p": 0,
            ".r": 0,
            ".end_kiss": 0,
            ".default_input_arrival": 0,
            ".default_output_required": 0,
            ".default_input_drive": 0,
            ".default_output_load": 0,
            ".default_max_input_load": 0,
            ".latch_order": 0,
            ".code": 0,
            ".exdc": 0
        }

    def __str__(self):
        """Printed string."""
        blif = self.model.__str__() + "\n"
        blif += self.inputs.__str__() + "\n"
        blif += self.outputs.__str__() + "\n"
        blif += "\n"

        if self.fsm.ispresent:
            blif += self.fsm.__str__() + "\n"
        else:
            for imported_file in self.imports:
                blif += imported_file.__str__() + "\n"

            for circuit in self.subcircuits:
                blif += circuit.__str__() + "\n"

            for latch in self.latches:
                blif += latch.__str__() + "\n"

            for function in self.booleanfunctions:
                blif += function.__str__() + "\n"

        blif += ".end\n"

        return blif



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Imports modules and packages for the user that executes:

python -m blifparser <input_path>
"""

from ._version import __version__

try:
    from . import keywords
    from . import utils
    from . import blifparser

except ImportError:
    from .blifparser import keywords
    from .blifparser import utils
    from .blifparser import blifparser

if __name__ == "__main__":
    blifparser.main()
